/*
 * SharedClause.hpp
 *
 *  Created on: 02.12.2009
 *      Author: kottlers
 */

#ifndef SHAREDCLAUSE_HPP_
#define SHAREDCLAUSE_HPP_
#define sharedclause_hpp_version "$Revision: 1.12 $"

#include "SharedBasic.hpp"
#include "GlobalSettings.hpp"

struct ClsLit{
	uint32_t has_next 	: 1;
	uint32_t propd		: 1;
	uint32_t lit		: 30;

	ClsLit(const Literal& l):has_next(1),propd(0),lit(l.idx()){asrt(Literal(lit)==l);}
	ClsLit(){}
};



class SharedClause{
	UserMask u_mask;
	uint8_t orig_lbd: 4;
	uint8_t wst		: 1; //can be removed
	uint8_t wst_soon: 1; //cannot be removed at the moment but can be ignored for var elimination
	uint8_t lrnt	: 1;
	uint8_t mrk		: 1;
	uint8_t pcnt;

	ClsLit lits[0];
public:
	static const uint16_t LBD_MAX =  15;
private:

	/**Construct a part of the clause.*/
	inline SharedClause(const UserMask& all_users, const uint32_t LBD, const bool learnt)
		:u_mask(all_users),orig_lbd(LBD<LBD_MAX?LBD:LBD_MAX),wst(0),wst_soon(0),lrnt(learnt),mrk(0),pcnt(0){}

	/**Copying of Clause instances is not allowed*/
	SharedClause(const SharedClause& cl):u_mask(cl.u_mask){asrt(false);}
	SharedClause& operator=(const SharedClause& c){asrt(false);return (SharedClause&)c;}

public:
	ClsLit* beg(){ return lits;}
	Literal getAt(const uint32_t& idx)const {return Literal(lits[idx].lit);}
	inline uint32_t countLits() const;

	static inline SharedClause* allocate(const UserMask& all_users, const uint32_t maxsz, const uint32_t lbd, const bool learnt);

	template<class B>
	friend 			class SolverLeader;
	friend			class Subsumption;
	bool 			isSoonWaste() const				{return wst_soon;}
	void 			setSoonWaste()					{wst_soon=1;}
public:


	void			setWaste()						{wst=1;} //only use this function when it's sure that all other threads have the replacement
	//inline void		setWaste(string why, int32_t val);

	bool 			isWaste() const					{return wst;}
	bool			isLearnt() const				{return lrnt;}
	bool			isFix() const					{return !isLearnt();}
	//void			changeToFix()					{lrnt = 0;} //may only be used when clause is not shared yet
	uint32_t		getOrigLBD() const					{asrt(isLearnt()|| orig_lbd==0);return orig_lbd;}
	UserMask		getUsers()const					{asrt(u_mask.hasAnyUser());return u_mask;}
	//inline bool		shrinkdLBD(uint32_t n);
	inline void		print();
	bool			hasUser(const UserBit& u)const	{return u_mask.hasUser(u);}
	uint32_t		getCount()const{				return pcnt;}
	inline void		incConflCnt();
	bool			isMrkd()const{					return mrk;}
	void			setMrk(bool b){					mrk=b;}
	//bool			mayContain(const Literal& l)const{return sign & hashVar(l);}
	//bool			containsLit(const Literal& l,uint32_t start=0)	const
	//												{return (bin_search(lits,start,sz,l) != max32);}

	//static inline void				setTrash(SharedClause* cls,const bool& immediate){
	//									if(immediate) cls->setWaste(); else cls->setRedundant();}
	template<class I>
	static inline SharedClause* 	construct(const UserMask& all_users, I lbeg, const I lend,const  uint32_t lbd,const bool lrnt);
	static inline SharedClause* 	copyFrom(const UserMask& all_users, const bool lrnt, SharedClause* from, const uint32_t& sz);
	static inline SharedClause* 	cloneSatDummyFrom(const UserMask& all_users, SharedClause* from, uint32_t from_sz);
	//static inline SharedClause*		copyRemv(const UserMask& all_users, const SharedClause* from, const uint32_t& new_sz,const Literal& skip);

	//static inline void				replace(SharedClause* old_cls, SharedClause* new_cls, const UserBit& ub);
	static void						evilDestruct(SharedClause* cls){free(cls);}

	/*static void						release(SharedClause* cls,const UserBit& u_bit, string why){asrt(cls!= NULL);
									LOG0("Release clause " << ((uint64_t)cls) << " because " << why);
										if(cls->u_mask.removeUser(u_bit)){ evilDestruct(cls);}}*/

	static void						release(SharedClause* cls,const UserBit& u_bit){asrt(cls!= NULL);
										if(cls->u_mask.removeUser(u_bit)){ evilDestruct(cls);}}


	//static inline bool				asrtLitsSorted(const SharedClause* cls);
};

#define FOR_LITS_STRT(cls,l) 	{ClsLit* clit = ((SharedClause*)cls)->beg(); do{ Literal l(clit->lit)
#define FOR_LITS_END 			}while((clit++)->has_next);}


/*
void	SharedClause::setWaste(string why, int32_t val){
	wst = 1;
#ifdef LOGGING_ACTUAL
	LOG0("WASTE clause at " << ((uint64_t)this)<<" with  reason: " << why << " " << val);
	if(isFix()){
		string c="";
		FOR_LITS_STRT(this,lit);
			char buff[25];
			sprintf(buff,"%d ",lit.toDIMACS());
			c += string(buff);
		FOR_LITS_END;

		LOG0("Lits are " << c);
	}
#endif
}
*/

/*
void 	SharedClause::setSoonWaste(){
	wst_soon = 1;
	LOG_BASIC("clause with id " << id << " set soon waste");
	if(isFix()){
		string c="";
		for(uint32_t i=0;i!=size();++i){
			char buff[25];
			sprintf(buff,"%d ",literals[i].toDIMACS());
			c += string(buff);
		}
		LOG_BASIC("lits are "<<c);
	}
}
*/


void	SharedClause::incConflCnt(){
	uint8_t count = pcnt;
	while(count < 255){
		if(exchange8(&pcnt,count,count+1)) return;
		count = pcnt;
	}
}

uint32_t SharedClause::countLits() const{
	uint32_t c = 0;
	ClsLit* clit = (ClsLit*)lits;
	do{++c;}while((clit++)->has_next);
	return c;
}

/*
bool	SharedClause::shrinkdLBD(uint32_t n){
	//asrt(n>1);
	if(n > MAX_LBD) return false;
	uint16_t curr = lbd;
	while(n<curr){
		if(exchange16(&lbd,curr,n)) return true;
		curr = lbd;}
	return false;
}*/


SharedClause* 	SharedClause::allocate(const UserMask& all_users, const uint32_t maxsz,const uint32_t lbd,const bool learnt){
	asrt(maxsz > 2); asrt(learnt || lbd==0); //asrt(!learnt || lbd>1);
	const uint32_t alloc_sz	= sizeof(SharedClause) + sizeof(ClsLit)*maxsz;
	void* mem				= malloc(alloc_sz);
	LOG3("Construct shared clause at " << (uint64_t)mem  << " with bytes " << alloc_sz);

	if(mem == NULL){
		LOG1("OUT OF MEMORY IN CLAUSE CREATION");
		fprintf(stderr,"\nc !!!OUT OF MEMORY!!!\n");
		exit(EXIT_FAILURE);
	}

	SharedClause* cls 	= new (mem) SharedClause(all_users,lbd,learnt);
 	return cls;
}




template<class I>
SharedClause* 	SharedClause::construct(const UserMask& all_users,I lbeg, const I lend, const uint32_t lbd, const bool lrnt){
	SharedClause* cls 	= allocate(all_users,lend-lbeg,((uint32_t)lrnt)*lbd,lrnt);
	ClsLit* clit		= cls->beg();
    for(;lbeg != lend;++lbeg) {
    	*clit++ = ClsLit(*lbeg);
    }
    asrt(clit > cls->beg());
    (--clit)->has_next = 0;  //asrt(asrtLitsSorted(cls));
	return cls;
}

/*
SharedClause*	SharedClause::copyRemv(const UserMask& all_users, const SharedClause* from, const uint32_t& new_sz,const Literal& skip){
	asrt(from->getOrigLBD()==0 || from->isLearnt());

	SharedClause* cls 	= allocate(all_users,new_sz,from->getOrigLBD(),from->isLearnt());

	ClsLit* flit		= ((SharedClause*)from)->beg();
	ClsLit* clit		= cls->beg();
	do{
		if(Literal(flit->lit) != skip){
			*(clit++)	= *flit;}

	}while((flit++)->has_next);

	(--clit)->has_next = 0;
	return cls;
}*/


SharedClause* 	SharedClause::cloneSatDummyFrom(const UserMask& all_users, SharedClause* from, uint32_t from_sz){
	const bool lrnt		= from->isLearnt();
	SharedClause* cls 	= allocate(all_users,from_sz + 1,((uint32_t)lrnt)*from->getOrigLBD(),lrnt);
	ClsLit* clit		= cls->beg();
	ClsLit* flit		= from->beg();
	*(clit++)			= ClsLit(TRUE_LIT);
	do{
		*(clit++)		= *flit;
	}while((flit++)->has_next);
	asrt((clit-1)->has_next == 0); //asrt(asrtLitsSorted(cls));

	//cls->setWaste("Cloned sat dummy and this is redundant sz ", from_sz);

	cls->setWaste();
	return cls;
}


SharedClause* 	SharedClause::copyFrom(const UserMask& all_users, const bool lrnt, SharedClause* from, const uint32_t& sz){
	SharedClause* cls 	= allocate(all_users,sz,((uint32_t)lrnt)*from->getOrigLBD(),lrnt);

	ClsLit* flit		= from->beg();
	ClsLit* clit		= cls->beg();
	do{
		*(clit++)		= *flit;
	}while((flit++)->has_next);
	asrt((flit-1)->has_next == 0);

	return cls;
}


/*
bool SharedClause::asrtLitsSorted(const SharedClause* cls) {
	Literal last(0);
	ClsLit* clit = ((SharedClause*)cls)->beg();
	do{
		Literal curr(clit->lit);
		if(last > curr){
			ASRT_PRNT("Clause is not sorted!!");
			((SharedClause*)cls)->print();
			return false;
		}
		last = curr;
	}while((clit++)->has_next);

	return true;
}
*/

void SharedClause::print() {

	string c="";
	FOR_LITS_STRT(this,lit);
		char buff[25];
		sprintf(buff,"%d ",lit.toDIMACS());
		c += string(buff);
	FOR_LITS_END;

	LOG1("Cls: " << c << " at loc " << ((uint64_t)(this)));
}

/*void SharedClause::replace(SharedClause* old_cls, SharedClause* new_cls, const UserBit& ub){
	asrt(!old_cls->hasUpdate());
	old_cls->linkUpdate(new_cls);
	forgetThis(old_cls,ub);
}


void SharedClause::forgetForever(SharedClause* cls,const UserBit& bit_id){
	asrt(cls != NULL); asrt(cls->hasUser(bit_id));
	forgetAll(cls,bit_id);
}*/

#endif /* SHAREDCLAUSE_HPP_ */
