
#ifndef SIBLINGINFO_HPP_
#define SIBLINGINFO_HPP_
#define siblinginfos_hpp_version      "$Revision: 1.6 $"
const char*  siblinginfos_cpp_version();

#include "GlobalSettings.hpp"
#include "LockFreeQueues.hpp"
#include "Satistics.hpp"
#include "SharedClause.hpp"
#include "VariableInfos.hpp"
#include "ElimClses.hpp"

struct LeaderSiblingInfo;
struct TrooperSiblingInfo;




#define MVAL asrt(type != TYPE_INVALID);
class Message{
	//not really graph updates but handled in same queue
	static const uint32_t TYPE_INVALID		= 0;
	static const uint32_t TYPE_UNIT_CLS		= 1;
	static const uint32_t TYPE_BIN_CLS		= 2;
	static const uint32_t TYPE_LONG_CLS		= 3;
	static const uint32_t TYPE_LIT_REPL		= 4;
	static const uint32_t TYPE_VAR_ELIM		= 5;
	static const uint32_t TYPE_BIN_REMV		= 6;
	//static const uint32_t TYPE_VAR_REMOVE	= 5;


	uint32_t	type 	: 3;
	uint32_t 	stuff 	: 29;
	uint64_t	data;
	Message(uint32_t t,uint32_t s, uint64_t i):	type(t),stuff(s),data(i){}
	Message(uint32_t t,uint32_t s, uint32_t i1, uint32_t i2):
		type(t),stuff(s),data(((uint64_t)i1) << 32 | (uint64_t)i2){}
	uint32_t 	splitDataFirst() const 	{return data >> 32;}
	uint32_t 	splitDataSecond() const {return data & max32;}
public:
	Message():type(TYPE_INVALID){};
	uint32_t getTypeCode() const	{return type;} //debugging

	static Message newUnitCls(const Literal& l)			{return Message(TYPE_UNIT_CLS,0,l.idx());}
	bool isUnitCls() const								{MVAL return type==TYPE_UNIT_CLS;}
	Literal getUnitCls()const							{MVAL asrt(isUnitCls());return Literal(data);}

	static Message newBinCls(const Literal& a,const Literal& b){
		return Message(TYPE_BIN_CLS,0,a.idx(),b.idx());}
	bool isBinCls() const								{MVAL return type==TYPE_BIN_CLS;}
	void getBinCls(Literal& a,Literal& b)const{
		asrt(isBinCls());a=Literal(splitDataFirst());b=Literal(splitDataSecond());}

	static Message newLongCls(const SharedClause* cls,const UserMask& recommend){
		return Message(TYPE_LONG_CLS,recommend.mask,(uint64_t)cls);}
	bool isLongCls() const								{MVAL return type==TYPE_LONG_CLS;}
	SharedClause* getLongCls(UserMask& recommended)const{
		asrt(isLongCls()); recommended.mask = stuff; return (SharedClause*)data;}

	static Message newLitRepl(const Literal& l,const Literal& repl){
		return Message(TYPE_LIT_REPL,0,l.idx(),repl.idx());}
	bool isLitRepl() const								{MVAL return type==TYPE_LIT_REPL;}
	Literal getLitRepl(Literal& repl)const{
		asrt(isLitRepl());repl=Literal(splitDataSecond());MVAL return Literal(splitDataFirst());}

	static Message newVarElim(const Variable& v){		return Message(TYPE_VAR_ELIM,0,v.idx());}
	bool isVarElim() const								{MVAL return type==TYPE_VAR_ELIM;}
	Variable getVarElim() const{
		asrt(isVarElim()); MVAL return Variable(data);}

	static Message newBinRemove(const Literal& a,const Literal& b){
		return Message(TYPE_BIN_REMV,0,a.idx(),b.idx());}
		bool isBinRemove() const						{MVAL return type==TYPE_BIN_REMV;}
		void getBinRemove(Literal& a,Literal& b)const{
			asrt(isBinRemove());a=Literal(splitDataFirst());b=Literal(splitDataSecond());}
};

//have to make it more dynamic
struct ImportCriteria{
	//uint32_t max_cls_sz;
	uint32_t force_after;
	uint32_t max_lbd_fact;
	uint32_t max_lbd_div;
	uint32_t max_true_permil;
	uint32_t lbd_mask;

	ImportCriteria(uint32_t force, bool take_orig_lbd, uint32_t max_lbd_factor,uint32_t max_lbd_divisor, uint32_t max_true_lits_permil):
		force_after(force), max_lbd_fact(max_lbd_factor),max_lbd_div(max_lbd_divisor),
		max_true_permil(max_true_lits_permil),lbd_mask(take_orig_lbd?0:SharedClause::LBD_MAX){}
};

/*
struct AutarkyGuess{
	uint32_t		chkd 	: 1;
	uint32_t 		sz		:31;
	AutarkyGuess* 	next;
	Literal			lits[0];
	AutarkyGuess(const uint32_t size):chkd(0),sz(size),next(NULL){}
};*/


struct SiblingInfo{

	typedef	GrowableMultReadQueue<Message>				queue_t;

	const UserID			user_id;
	const UserBit			user_bit;
	const UserMask			all_users_bits;
	const UserMask			all_but_me_bits;
	//AutarkyGuess*			aut_last; //for writing thread
	//volatile AutarkyGuess*	aut_next; //for reading thread (leader)

protected:
	uint32_t				last_long_pop;
	queue_t					stream;
	vector<SiblingInfo*> 	siblings;


public:
	Satistics*				stats;
	BinVarInfo*				v_info;
	ElimClses&				elim;
	uint32_t				max_lbd_gc;
	ImportCriteria 			import_criteria;


	uint32_t		innerGroupSize()const	{return siblings.size()+1;}
	uint32_t 		receiverCount() const	{return siblings.size();}

	uint32_t 		getMaxImportLBD()const 	{return (max_lbd_gc*import_criteria.max_lbd_fact)/import_criteria.max_lbd_div;}
	uint32_t 		getMaxTrueLitsPermil()const{return import_criteria.max_true_permil;}
	uint32_t		getImportFreqForced()const{return import_criteria.force_after;}
	uint32_t		getImportLbdMsk() const{ return import_criteria.lbd_mask;}
	//void popLongs(vector<SharedClause*>& out);
	//SharedClause* popLong();

	//void setLastLBD(const uint32_t lbd){max_lbd_gc = lbd;}
	void setLastLBD(const uint32_t lbd){set_max_int(max_lbd_gc,lbd);}

	SiblingInfo(Satistics* s,uint32_t grp_sz,uint32_t thread_id,ImportCriteria imprt,ElimClses& eli);

	~SiblingInfo();
	void initVarInfo(BinVarInfo* vinf){v_info=vinf;}

	bool	popMessage(Message& m);
	Literal getMayorValue(const Variable& v);
	void	pushAutarkyGuess(Literal lits[], const uint32_t sz);
	//AutarkyGuess* popAutarkyGuess();
protected:

	void	sendUnitCls(const Literal& lit);
	void	sendBinCls(const Literal& a, const Literal& b);
	void	sendLongCls(const SharedClause* cls,const UserMask& recommended_for);

	void 	deleteQueueContent();
	void 	addSibling(SiblingInfo* q);

};


class TrooperSiblingInfo: public SiblingInfo{
	friend class LeaderSiblingInfo;
public:

	TrooperSiblingInfo(Satistics* s,uint32_t grp_sz,uint32_t thread_id,const UserBit& leader_ubit,ImportCriteria imprt,ElimClses& eli);
	~TrooperSiblingInfo();

	void	sendUnitCls(const Literal& lit){
			SiblingInfo::sendUnitCls(lit);}
	void	sendBinCls(const Literal& a, const Literal& b){
			SiblingInfo::sendBinCls(a,b);}
	void	sendLongCls(const SharedClause* cls,const UserMask& recommended_for){
			SiblingInfo::sendLongCls(cls,recommended_for);}

	void deleteQueueContent(){SiblingInfo::deleteQueueContent();}

	void addTrooper(TrooperSiblingInfo* trooper);
	void addLeader(LeaderSiblingInfo* leader);
};


class LeaderSiblingInfo: public SiblingInfo{

	friend class TrooperSiblingInfo;

	//vector<TrooperSiblingInfo*> troopers;
	bool						active;

public:
	LeaderSiblingInfo(Satistics* s,uint32_t grp_sz,uint32_t thread_id,ImportCriteria imprt,ElimClses& eli);

	~LeaderSiblingInfo();

	void activate(){active = true;}
	bool isActive() const {return active;}

	bool single() const	{return all_users_bits.isOnlyUser(user_bit);}

	//void send(SharedClause* cls);

	//GraphUpdateUnit& startGUSending(const UserID& who,uint16_t type);

	void	sendUnitCls(const Literal& lit){
			if(active)SiblingInfo::sendUnitCls(lit);}
	void	sendBinCls(const Literal& a, const Literal& b){
			if(active)SiblingInfo::sendBinCls(a,b);}
	void	sendLongCls(const SharedClause* cls,const UserMask& recommended_for){
			if(active)SiblingInfo::sendLongCls(cls,recommended_for);}
	void	sendMessages(const vector<Message>& mess){
			if(active)SiblingInfo::stream.push(mess);}
	void	sendMessage(const Message& mess){
			if(active)SiblingInfo::stream.push(mess);}


	//bool popShort(ShortCls& cls,UserID& s_id);

	void printParaProgress();
	/*uint32_t getStartsOfTrooper(){
		return troopers.empty()?0:troopers[0]->stats->nbr_starts;}*/

	void deleteQueueContent();

	void addTrooper(TrooperSiblingInfo* trooper);

	uint32_t getMaxLBDThreash();

	//AutarkyGuess* popNextAutarkyGuess();
};




#endif /* SIBLINGINFO_HPP_ */
